'use strict';
Object.defineProperty(exports, '__esModule', { value: true });
exports.setTrackUpdate = exports.finishCommitWork = exports.applyCommit = exports.createNativeElement = void 0;
const core_1 = require('@dark-engine/core');
const portal_1 = require('../portal');
const events_1 = require('../events');
const constants_1 = require('../constants');
const attrBlackListMap = {
  [core_1.ATTR_KEY]: true,
  [core_1.ATTR_REF]: true,
  [core_1.ATTR_FLAG]: true,
};
let fragmentsMap = new Map();
let moves = [];
let trackUpdate = null;
const svgTagNamesMap = (0, core_1.keyBy)(constants_1.SVG_TAG_NAMES.split(','), x => x);
const voidTagNamesMap = (0, core_1.keyBy)(constants_1.VOID_TAG_NAMES.split(','), x => x);
const createNativeElementMap = {
  [core_1.NodeType.TAG]: vNode => {
    const tagNode = vNode;
    const node = detectIsSvgElement(tagNode.name)
      ? document.createElementNS('http://www.w3.org/2000/svg', tagNode.name)
      : document.createElement(tagNode.name);
    return node;
  },
  [core_1.NodeType.TEXT]: vNode => {
    const textNode = vNode;
    const node = document.createTextNode(textNode.value);
    return node;
  },
  [core_1.NodeType.COMMENT]: vNode => {
    const commentNode = vNode;
    const node = document.createComment(commentNode.value);
    return node;
  },
};
function createNativeElement(vNode) {
  return createNativeElementMap[vNode.type](vNode);
}
exports.createNativeElement = createNativeElement;
function detectIsSvgElement(tagName) {
  return Boolean(svgTagNamesMap[tagName]);
}
function detectIsVoidElement(tagName) {
  return Boolean(voidTagNamesMap[tagName]);
}
function applyRef(ref, element) {
  (0, core_1.applyRef)(ref, element);
}
function addAttributes(element, vNode) {
  if (!(0, core_1.detectIsTagVirtualNode)(vNode)) return;
  const attrNames = Object.keys(vNode.attrs);
  const tagElement = element;
  for (const attrName of attrNames) {
    const attrValue = vNode.attrs[attrName];
    if (attrName === core_1.ATTR_REF) {
      applyRef(attrValue, element);
      continue;
    }
    if ((0, core_1.detectIsFunction)(attrValue)) {
      if ((0, events_1.detectIsEvent)(attrName)) {
        (0, events_1.delegateEvent)({
          target: tagElement,
          handler: attrValue,
          eventName: (0, events_1.getEventName)(attrName),
        });
      }
    } else if (!(0, core_1.detectIsUndefined)(attrValue) && !attrBlackListMap[attrName]) {
      const stop = patchProperties({
        tagName: vNode.name,
        element: tagElement,
        attrValue,
        attrName,
      });
      !stop && tagElement.setAttribute(attrName, attrValue);
    }
  }
}
function updateAttributes(element, vNode, nextVNode) {
  const attrNames = new Set([...Object.keys(vNode.attrs), ...Object.keys(nextVNode.attrs)]);
  const tagElement = element;
  for (const attrName of attrNames) {
    const prevAttrValue = vNode.attrs[attrName];
    const nextAttrValue = nextVNode.attrs[attrName];
    if (attrName === core_1.ATTR_REF) {
      applyRef(prevAttrValue, element);
      continue;
    }
    if (!(0, core_1.detectIsUndefined)(nextAttrValue)) {
      if ((0, core_1.detectIsFunction)(prevAttrValue)) {
        if ((0, events_1.detectIsEvent)(attrName) && prevAttrValue !== nextAttrValue) {
          (0, events_1.delegateEvent)({
            target: tagElement,
            handler: nextAttrValue,
            eventName: (0, events_1.getEventName)(attrName),
          });
        }
      } else if (!attrBlackListMap[attrName] && prevAttrValue !== nextAttrValue) {
        const stop = patchProperties({
          tagName: nextVNode.name,
          element: tagElement,
          attrValue: nextAttrValue,
          attrName,
        });
        !stop && tagElement.setAttribute(attrName, nextAttrValue);
      }
    } else {
      tagElement.removeAttribute(attrName);
    }
  }
}
function patchProperties(options) {
  const { tagName, element, attrName, attrValue } = options;
  const fn = patchPropertiesSpecialCasesMap[tagName];
  let stop = fn ? fn(element, attrName, attrValue) : false;
  if (canSetProperty(Object.getPrototypeOf(element), attrName)) {
    element[attrName] = attrValue;
  }
  if (!stop && (0, core_1.detectIsBoolean)(attrValue)) {
    stop = !attrName.includes('-');
  }
  return stop;
}
function canSetProperty(prototype, key) {
  return prototype.hasOwnProperty(key) && Boolean(Object.getOwnPropertyDescriptor(prototype, key)?.set);
}
const patchPropertiesSpecialCasesMap = {
  input: (element, attrName, attrValue) => {
    if (attrName === 'value' && (0, core_1.detectIsBoolean)(attrValue)) {
      element.checked = attrValue;
    } else if (attrName === 'autoFocus') {
      element.autofocus = Boolean(attrValue);
    }
    return false;
  },
  textarea: (element, attrName, attrValue) => {
    if (attrName === 'value') {
      element.innerHTML = String(attrValue);
      return true;
    }
    return false;
  },
};
function getParentFiberWithNativeElement(fiber) {
  let nextFiber = fiber;
  while (nextFiber) {
    nextFiber = nextFiber.parent;
    if ((0, portal_1.detectIsPortal)(nextFiber.instance)) {
      nextFiber.nativeElement = (0, portal_1.getPortalContainer)(nextFiber.instance);
    }
    if (nextFiber.nativeElement) return nextFiber;
  }
  return nextFiber;
}
function append(fiber, parentElement) {
  const { fragment } = fragmentsMap.get(parentElement) || {
    fragment: document.createDocumentFragment(),
    callback: () => {},
  };
  fragmentsMap.set(parentElement, {
    fragment,
    callback: () => {
      parentElement.appendChild(fragment);
    },
  });
  fragment.appendChild(fiber.nativeElement);
}
function insert(fiber, parentElement) {
  parentElement.insertBefore(fiber.nativeElement, parentElement.childNodes[fiber.elementIdx]);
}
function commitCreation(fiber) {
  const parentFiber = getParentFiberWithNativeElement(fiber);
  const parentElement = parentFiber.nativeElement;
  const childNodes = parentElement.childNodes;
  if (core_1.isHydrateZone.get()) {
    const nativeElement = childNodes[fiber.elementIdx];
    if (
      (0, core_1.detectIsTextVirtualNode)(fiber.instance) &&
      nativeElement instanceof Text &&
      fiber.instance.value.length !== nativeElement.length
    ) {
      nativeElement.splitText(fiber.instance.value.length);
    }
    fiber.nativeElement = nativeElement;
  } else {
    if (childNodes.length === 0 || fiber.elementIdx > childNodes.length - 1) {
      const vNode = parentFiber.instance;
      !detectIsVoidElement(vNode.name) && append(fiber, parentElement);
    } else {
      insert(fiber, parentElement);
    }
  }
  addAttributes(fiber.nativeElement, fiber.instance);
}
function commitUpdate(fiber) {
  const element = fiber.nativeElement;
  const prevInstance = fiber.alternate.instance;
  const nextInstance = fiber.instance;
  if (
    (0, core_1.detectIsTextVirtualNode)(prevInstance) &&
    (0, core_1.detectIsTextVirtualNode)(nextInstance) &&
    prevInstance.value !== nextInstance.value
  ) {
    return (element.textContent = nextInstance.value);
  }
  if ((0, core_1.detectIsTagVirtualNode)(prevInstance) && (0, core_1.detectIsTagVirtualNode)(nextInstance)) {
    return updateAttributes(element, prevInstance, nextInstance);
  }
}
function commitDeletion(fiber) {
  const parentFiber = getParentFiberWithNativeElement(fiber);
  (0, core_1.walkFiber)(fiber, ({ nextFiber, isReturn, resetIsDeepWalking, stop }) => {
    if (nextFiber === fiber.nextSibling || nextFiber === fiber.parent) {
      return stop();
    }
    if (!isReturn && nextFiber.nativeElement) {
      !(0, portal_1.detectIsPortal)(nextFiber.instance) &&
        parentFiber.nativeElement.removeChild(nextFiber.nativeElement);
      return resetIsDeepWalking();
    }
  });
}
function move(fiber) {
  const sourceNodes = collectElements(fiber);
  const sourceNode = sourceNodes[0];
  const parentElement = sourceNode.parentElement;
  const sourceFragment = new DocumentFragment();
  const elementIdx = fiber.elementIdx;
  let idx = 0;
  const move = () => {
    for (let i = 1; i < sourceNodes.length; i++) {
      parentElement.removeChild(parentElement.childNodes[elementIdx + 1]);
    }
    parentElement.replaceChild(sourceFragment, parentElement.childNodes[elementIdx]);
  };
  for (const node of sourceNodes) {
    parentElement.insertBefore(document.createComment(`${elementIdx}:${idx}`), node);
    sourceFragment.appendChild(node);
    idx++;
  }
  moves.push(move);
}
function collectElements(fiber) {
  const store = [];
  (0, core_1.walkFiber)(fiber, ({ nextFiber, isReturn, resetIsDeepWalking, stop }) => {
    if (nextFiber === fiber.nextSibling || nextFiber === fiber.parent) {
      return stop();
    }
    if (!isReturn && nextFiber.nativeElement) {
      !(0, portal_1.detectIsPortal)(nextFiber.instance) && store.push(nextFiber.nativeElement);
      return resetIsDeepWalking();
    }
  });
  return store;
}
const applyCommitMap = {
  [core_1.EffectTag.CREATE]: fiber => {
    if (fiber.nativeElement === null) return;
    trackUpdate && trackUpdate(fiber.nativeElement);
    commitCreation(fiber);
  },
  [core_1.EffectTag.UPDATE]: fiber => {
    if (fiber.move) {
      move(fiber);
      fiber.move = false;
    }
    if (
      fiber.nativeElement === null ||
      !(0, core_1.detectIsVirtualNode)(fiber.alternate.instance) ||
      !(0, core_1.detectIsVirtualNode)(fiber.instance)
    ) {
      return;
    }
    trackUpdate && trackUpdate(fiber.nativeElement);
    commitUpdate(fiber);
  },
  [core_1.EffectTag.DELETE]: fiber => commitDeletion(fiber),
  [core_1.EffectTag.SKIP]: () => {},
};
function applyCommit(fiber) {
  applyCommitMap[fiber.effectTag](fiber);
}
exports.applyCommit = applyCommit;
function finishCommitWork() {
  for (const { callback } of fragmentsMap.values()) {
    callback();
  }
  for (const move of moves) {
    move();
  }
  fragmentsMap = new Map();
  moves = [];
  core_1.isHydrateZone.set(false);
}
exports.finishCommitWork = finishCommitWork;
function setTrackUpdate(fn) {
  trackUpdate = fn;
}
exports.setTrackUpdate = setTrackUpdate;
//# sourceMappingURL=dom.js.map
