import {
  ATTR_KEY,
  ATTR_REF,
  ATTR_FLAG,
  EffectTag,
  detectIsFunction,
  detectIsUndefined,
  detectIsBoolean,
  keyBy,
  NodeType,
  detectIsVirtualNode,
  detectIsTagVirtualNode,
  detectIsTextVirtualNode,
  walkFiber,
  isHydrateZone,
  applyRef as applyRef$,
} from '@dark-engine/core';
import { detectIsPortal, getPortalContainer } from '../portal';
import { delegateEvent, detectIsEvent, getEventName } from '../events';
import { SVG_TAG_NAMES, VOID_TAG_NAMES } from '../constants';
const attrBlackListMap = {
  [ATTR_KEY]: true,
  [ATTR_REF]: true,
  [ATTR_FLAG]: true,
};
let fragmentsMap = new Map();
let moves = [];
let trackUpdate = null;
const svgTagNamesMap = keyBy(SVG_TAG_NAMES.split(','), x => x);
const voidTagNamesMap = keyBy(VOID_TAG_NAMES.split(','), x => x);
const createNativeElementMap = {
  [NodeType.TAG]: vNode => {
    const tagNode = vNode;
    const node = detectIsSvgElement(tagNode.name)
      ? document.createElementNS('http://www.w3.org/2000/svg', tagNode.name)
      : document.createElement(tagNode.name);
    return node;
  },
  [NodeType.TEXT]: vNode => {
    const textNode = vNode;
    const node = document.createTextNode(textNode.value);
    return node;
  },
  [NodeType.COMMENT]: vNode => {
    const commentNode = vNode;
    const node = document.createComment(commentNode.value);
    return node;
  },
};
function createNativeElement(vNode) {
  return createNativeElementMap[vNode.type](vNode);
}
function detectIsSvgElement(tagName) {
  return Boolean(svgTagNamesMap[tagName]);
}
function detectIsVoidElement(tagName) {
  return Boolean(voidTagNamesMap[tagName]);
}
function applyRef(ref, element) {
  applyRef$(ref, element);
}
function addAttributes(element, vNode) {
  if (!detectIsTagVirtualNode(vNode)) return;
  const attrNames = Object.keys(vNode.attrs);
  const tagElement = element;
  for (const attrName of attrNames) {
    const attrValue = vNode.attrs[attrName];
    if (attrName === ATTR_REF) {
      applyRef(attrValue, element);
      continue;
    }
    if (detectIsFunction(attrValue)) {
      if (detectIsEvent(attrName)) {
        delegateEvent({
          target: tagElement,
          handler: attrValue,
          eventName: getEventName(attrName),
        });
      }
    } else if (!detectIsUndefined(attrValue) && !attrBlackListMap[attrName]) {
      const stop = patchProperties({
        tagName: vNode.name,
        element: tagElement,
        attrValue,
        attrName,
      });
      !stop && tagElement.setAttribute(attrName, attrValue);
    }
  }
}
function updateAttributes(element, vNode, nextVNode) {
  const attrNames = new Set([...Object.keys(vNode.attrs), ...Object.keys(nextVNode.attrs)]);
  const tagElement = element;
  for (const attrName of attrNames) {
    const prevAttrValue = vNode.attrs[attrName];
    const nextAttrValue = nextVNode.attrs[attrName];
    if (attrName === ATTR_REF) {
      applyRef(prevAttrValue, element);
      continue;
    }
    if (!detectIsUndefined(nextAttrValue)) {
      if (detectIsFunction(prevAttrValue)) {
        if (detectIsEvent(attrName) && prevAttrValue !== nextAttrValue) {
          delegateEvent({
            target: tagElement,
            handler: nextAttrValue,
            eventName: getEventName(attrName),
          });
        }
      } else if (!attrBlackListMap[attrName] && prevAttrValue !== nextAttrValue) {
        const stop = patchProperties({
          tagName: nextVNode.name,
          element: tagElement,
          attrValue: nextAttrValue,
          attrName,
        });
        !stop && tagElement.setAttribute(attrName, nextAttrValue);
      }
    } else {
      tagElement.removeAttribute(attrName);
    }
  }
}
function patchProperties(options) {
  const { tagName, element, attrName, attrValue } = options;
  const fn = patchPropertiesSpecialCasesMap[tagName];
  let stop = fn ? fn(element, attrName, attrValue) : false;
  if (canSetProperty(Object.getPrototypeOf(element), attrName)) {
    element[attrName] = attrValue;
  }
  if (!stop && detectIsBoolean(attrValue)) {
    stop = !attrName.includes('-');
  }
  return stop;
}
function canSetProperty(prototype, key) {
  return prototype.hasOwnProperty(key) && Boolean(Object.getOwnPropertyDescriptor(prototype, key)?.set);
}
const patchPropertiesSpecialCasesMap = {
  input: (element, attrName, attrValue) => {
    if (attrName === 'value' && detectIsBoolean(attrValue)) {
      element.checked = attrValue;
    } else if (attrName === 'autoFocus') {
      element.autofocus = Boolean(attrValue);
    }
    return false;
  },
  textarea: (element, attrName, attrValue) => {
    if (attrName === 'value') {
      element.innerHTML = String(attrValue);
      return true;
    }
    return false;
  },
};
function getParentFiberWithNativeElement(fiber) {
  let nextFiber = fiber;
  while (nextFiber) {
    nextFiber = nextFiber.parent;
    if (detectIsPortal(nextFiber.instance)) {
      nextFiber.nativeElement = getPortalContainer(nextFiber.instance);
    }
    if (nextFiber.nativeElement) return nextFiber;
  }
  return nextFiber;
}
function append(fiber, parentElement) {
  const { fragment } = fragmentsMap.get(parentElement) || {
    fragment: document.createDocumentFragment(),
    callback: () => {},
  };
  fragmentsMap.set(parentElement, {
    fragment,
    callback: () => {
      parentElement.appendChild(fragment);
    },
  });
  fragment.appendChild(fiber.nativeElement);
}
function insert(fiber, parentElement) {
  parentElement.insertBefore(fiber.nativeElement, parentElement.childNodes[fiber.elementIdx]);
}
function commitCreation(fiber) {
  const parentFiber = getParentFiberWithNativeElement(fiber);
  const parentElement = parentFiber.nativeElement;
  const childNodes = parentElement.childNodes;
  if (isHydrateZone.get()) {
    const nativeElement = childNodes[fiber.elementIdx];
    if (
      detectIsTextVirtualNode(fiber.instance) &&
      nativeElement instanceof Text &&
      fiber.instance.value.length !== nativeElement.length
    ) {
      nativeElement.splitText(fiber.instance.value.length);
    }
    fiber.nativeElement = nativeElement;
  } else {
    if (childNodes.length === 0 || fiber.elementIdx > childNodes.length - 1) {
      const vNode = parentFiber.instance;
      !detectIsVoidElement(vNode.name) && append(fiber, parentElement);
    } else {
      insert(fiber, parentElement);
    }
  }
  addAttributes(fiber.nativeElement, fiber.instance);
}
function commitUpdate(fiber) {
  const element = fiber.nativeElement;
  const prevInstance = fiber.alternate.instance;
  const nextInstance = fiber.instance;
  if (
    detectIsTextVirtualNode(prevInstance) &&
    detectIsTextVirtualNode(nextInstance) &&
    prevInstance.value !== nextInstance.value
  ) {
    return (element.textContent = nextInstance.value);
  }
  if (detectIsTagVirtualNode(prevInstance) && detectIsTagVirtualNode(nextInstance)) {
    return updateAttributes(element, prevInstance, nextInstance);
  }
}
function commitDeletion(fiber) {
  const parentFiber = getParentFiberWithNativeElement(fiber);
  walkFiber(fiber, ({ nextFiber, isReturn, resetIsDeepWalking, stop }) => {
    if (nextFiber === fiber.nextSibling || nextFiber === fiber.parent) {
      return stop();
    }
    if (!isReturn && nextFiber.nativeElement) {
      !detectIsPortal(nextFiber.instance) && parentFiber.nativeElement.removeChild(nextFiber.nativeElement);
      return resetIsDeepWalking();
    }
  });
}
function move(fiber) {
  const sourceNodes = collectElements(fiber);
  const sourceNode = sourceNodes[0];
  const parentElement = sourceNode.parentElement;
  const sourceFragment = new DocumentFragment();
  const elementIdx = fiber.elementIdx;
  let idx = 0;
  const move = () => {
    for (let i = 1; i < sourceNodes.length; i++) {
      parentElement.removeChild(parentElement.childNodes[elementIdx + 1]);
    }
    parentElement.replaceChild(sourceFragment, parentElement.childNodes[elementIdx]);
  };
  for (const node of sourceNodes) {
    parentElement.insertBefore(document.createComment(`${elementIdx}:${idx}`), node);
    sourceFragment.appendChild(node);
    idx++;
  }
  moves.push(move);
}
function collectElements(fiber) {
  const store = [];
  walkFiber(fiber, ({ nextFiber, isReturn, resetIsDeepWalking, stop }) => {
    if (nextFiber === fiber.nextSibling || nextFiber === fiber.parent) {
      return stop();
    }
    if (!isReturn && nextFiber.nativeElement) {
      !detectIsPortal(nextFiber.instance) && store.push(nextFiber.nativeElement);
      return resetIsDeepWalking();
    }
  });
  return store;
}
const applyCommitMap = {
  [EffectTag.CREATE]: fiber => {
    if (fiber.nativeElement === null) return;
    trackUpdate && trackUpdate(fiber.nativeElement);
    commitCreation(fiber);
  },
  [EffectTag.UPDATE]: fiber => {
    if (fiber.move) {
      move(fiber);
      fiber.move = false;
    }
    if (
      fiber.nativeElement === null ||
      !detectIsVirtualNode(fiber.alternate.instance) ||
      !detectIsVirtualNode(fiber.instance)
    ) {
      return;
    }
    trackUpdate && trackUpdate(fiber.nativeElement);
    commitUpdate(fiber);
  },
  [EffectTag.DELETE]: fiber => commitDeletion(fiber),
  [EffectTag.SKIP]: () => {},
};
function applyCommit(fiber) {
  applyCommitMap[fiber.effectTag](fiber);
}
function finishCommitWork() {
  for (const { callback } of fragmentsMap.values()) {
    callback();
  }
  for (const move of moves) {
    move();
  }
  fragmentsMap = new Map();
  moves = [];
  isHydrateZone.set(false);
}
function setTrackUpdate(fn) {
  trackUpdate = fn;
}
export { createNativeElement, applyCommit, finishCommitWork, setTrackUpdate };
//# sourceMappingURL=dom.js.map
